# -*- coding: iso-8859-1 -*-
#
# $Id: MultipleChoiceTest.py,v 1.24 2005/10/21 13:46:13 mxp Exp $
#
# Copyright  2004 Otto-von-Guericke-Universitt Magdeburg
#
# This file is part of LlsMultipleChoice.
#
# LlsMultipleChoice is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LlsMultipleChoice is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LlsMultipleChoice; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, os.path
from AccessControl import ClassSecurityInfo, getSecurityManager
from Acquisition import aq_base, aq_acquire, aq_inner, aq_parent
from DateTime import DateTime
from zipfile import ZipFile
from Products.Archetypes.public import Schema, \
    BooleanField, IntegerField, ObjectField, LinesField, registerType, \
    FileField
from Products.Archetypes.Widget import TypesWidget, BooleanWidget, \
    IntegerWidget, FileWidget
from Products.CMFCore.utils import getToolByName
try: # Plone 2.0.5
    from Products.CMFPlone.PloneUtilities import localized_time
except: # Plone 2.1
    from Products.CMFPlone.utils import ulocalized_time as localized_time
from Products.LlsMultipleChoice.config import PROJECTNAME, SKINS_DIR, \
    QUESTION_DIR, QUESTION_TYPES, ANSWER_DIR, \
    ANSWER_TYPES, GLOBALS, PERMISSION_INTERROGATOR, PERMISSION_STUDENT, \
    I18N_DOMAIN, CUSTOM_EVALUATION_FUNCTION_NAME
from Products.LlsMultipleChoice.AbstractQuestionContainer import AbstractQuestionContainer
from Products.LlsMultipleChoice.QuestionTypes.MultipleChoiceQuestion import MultipleChoiceQuestion
from Products.LlsMultipleChoice.QuestionGroup import QuestionGroup
from Products.LlsMultipleChoice.tools import getParent, evalFunString, \
    log, isNumeric, MyStringIO, makeTransactionUnundoable
from Products.LlsMultipleChoice.qti import importAssessmentItem, importPackage, exportPackage


class ResultsWidget(TypesWidget):
    """ A custom widget for handling a dictionary of 
        CandidateTestResult objects. 
    """
    _properties = TypesWidget._properties.copy()
    _properties.update({
        # The Page Template for this widget
        'macro' : 'results_widget',
    })
    
    
class EvaluationScriptsWidget(TypesWidget):
    """ A custom widget for handling the 'evaluationScripts' 
        property of the MultipleChoiceTest.
    """
    _properties = TypesWidget._properties.copy()
    _properties.update({
        'macro' : 'evaluation_scripts_widget',
    })


class MultipleChoiceTest(AbstractQuestionContainer):
    """ This is the main class of the 'LlsMultipleChoice' Product.
        A multiple choice test is basically a folder that contains 
        questions and groups of questions, i.e. folders that
        contain questions.
    """
    
    # format: [file extension, column delimiter, row delimiter, 
    # start of text delimiter, end of text delimiter, escape char]
    RESULTS_EXPORT_FORMAT_TAB = ['tab', '\t', '\n', '"', '"', '"']
    RESULTS_EXPORT_FORMAT_CSV = ['csv', ';',  '\n', '"', '"', '"']
    RESULTS_EXPORT_FORMATS = [RESULTS_EXPORT_FORMAT_TAB, RESULTS_EXPORT_FORMAT_CSV]
    
    # See comments in 'AbstractQuestionContainer' for an explanation
    # of the function of a schema, therein defined properties (fields) and 
    # internationalization of the widgets
    schema = AbstractQuestionContainer.schema + Schema((
            ObjectField('results',
                        # A dictionary that contains CandidateTestResult objects.
                        # The keys are the candidate IDs. The value for a key is the corresponding 
                        # CandidateTestResult instance. (That means that you cannot save
                        # more than one CandidateTestResult instance for a candidate.
                        required=False,
                        # Not sure what this does.
                        multiValued=True,
                        default={},
                        mutator='setResults',
                        widget=ResultsWidget(),
                        read_permission=PERMISSION_INTERROGATOR,
            ),
            BooleanField('instantFeedback',
                        # Instant feedback means that the candidate will see their results 
                        # immediately after submitting their test.
                        required=False,
                        default=0,
                        accessor='isInstantFeedback',
                        widget=BooleanWidget(
                            label='Instant Feedback',
                            label_msgid='instant_feedback_label',
                            description='If you want to give the candidates instant feedback '\
                                'check this box.',
                            description_msgid='instant_feedback_tool_tip',
                            i18n_domain=I18N_DOMAIN),
            ),
            BooleanField('allowRepetition',
                        # See 'description' of the widget.
                        required=False,
                        default=0,
                        accessor='isAllowRepetition',
                        widget=BooleanWidget(
                            label='Allow Repetition',
                            label_msgid='allow_repetition_label',
                            description='If you want to allow repeated submission of the test '\
                                'check this box.',
                            description_msgid='allow_repetition_tool_tip',
                            i18n_domain=I18N_DOMAIN),
            ),
            ObjectField('evaluationScripts',
                        # A dictionary for custom evaluation scripts. It is possible to 
                        # customize the evaluation behavior of the LlsMultipleChoice Product by 
                        # uploading custom evaluation scripts for 1) the MultipleChoiceTest, 
                        # 2) QuestionGroup objects, 3) Question objects. Otherwise a default 
                        # evaluation method will be used.
                        #   The keys of this dictionary are the 'archetype_name' properties of 
                        # those classes. The values will be the scripts.
                        required=False,
                        # Not sure what this does.
                        multiValued=True,
                        default={},
                        mutator='setEvaluationScripts',
                        widget=EvaluationScriptsWidget(),
                        read_permission=PERMISSION_INTERROGATOR,
            ),
        ),)

    """ Inherit custom actions from 'AbstractQuestionContainer' and redefine them and/or 
        define some new ones. 
    """
    actions = AbstractQuestionContainer.actions + (
        {
            'id'           : 'view',
            'action'       : 'string:${object_url}/test_view',
            'permissions'  : (PERMISSION_STUDENT,),
        },
        {
            'id'           : 'upload_evaluation_scripts',
            'name'         : 'Script Upload',
            'action'       : 'string:${object_url}/upload_evaluation_scripts',
            'permissions'  : (PERMISSION_INTERROGATOR,),
        },
        {
            'id'           : 'test_import_export',
            'name'         : 'Import/Export',
            'action'       : 'string:${object_url}/test_import_export',
            'permissions'  : (PERMISSION_INTERROGATOR,),
        },
        {
            'id'           : 'see_results',
            'name'         : 'Results',
            'action'       : 'string:${object_url}/test_results',
            'permissions'  : (PERMISSION_INTERROGATOR,),
        },
    )
    
    meta_type = portal_type = archetype_name = 'MC Test'
    # This type is directly allowed anywhere.
    global_allow = True
    # MultipleChoiceTests may only contain questions and groups of question.
    allowed_content_types = (QuestionGroup.archetype_name, 'Folder', 'File', 'Image') \
        + AbstractQuestionContainer.allowed_content_types
    #~ log("allowed_content_types = " + str(allowed_content_types) + "\n")
    """ A custom icon for MultipleChoiceTests. The icon is located in 
        the skins directory ('skins/llsMultipleChoice') (It is possible to put it
        in another folder but the path must be known to Zope.). 
    """
    content_icon = 'MultipleChoiceTest.png'
    
    security = ClassSecurityInfo()
    
    
    def __getattr__(self, key):
        """ This workaround method maps  read requests for the standard 
            Plone properties 'effective_date'  and 'expiration_date' to
            calls to  the Archetype  methods  'getEffectiveDate()'  and 
            'getExpirationDate()',  respectively. This way timed publi-
            shing/unpublishing works with MultipleChoiceTests, too.
              Consult  the Python  documentation  to  learn  more about
            '__getattr__()' methods.
        """
        if(key == 'effective_date'):
            return self.getEffectiveDate()
        elif(key == 'expiration_date'):
            return self.getExpirationDate()
        else:
            raise AttributeError("'%s' object has no attribute '%s'" %(repr(self), str(key)) )

            
    def hasParticipated(self, candidateId):
        """ Find out if a test has been generated for candidate "candidateId", 
            if he has broken the seal, so to speak. 
        """
        return self.getResults().has_key(candidateId)
        
    
    #~ security.declareProtected(PERMISSION_INTERROGATOR, 'getParticipantIds')
    def getParticipantIds(self):
        """ Get the IDs of all the candidates for whom a test has been generated. """
        return self.getResults().keys()
        
        
    def hasSubmitted(self, candidateId):
        """ Find out if a candidate has actually submitted/taken this test. """
        #~ log("MultipleChoiceTest.hasSubmitted():\n")
        if(self.results.has_key(candidateId)):
            #~ log("\tself.results.has_key(" + candidateId + ")\n")
            #~ log("\t(self.getCandidateTimeFinish(" + candidateId + ") != None) == " \
              #~ + str(self.getCandidateTimeFinish(    candidateId    ) != None) + "\n")
            return (self.getCandidateTimeFinish(candidateId) != None)
        else:
            #~ log("\tnot self.results.has_key(" + candidateId + ")\n")
            return False
            
    
    def getCandidateTimeStart(self, candidateId):
        """ Returns the time when the candidate opened his/her test.
            (The start time is set in AbstractQuestionContainer.setAnswers())
        """
        #~ log('MultipleChoiceTest.getCandidateTimeStart()\n')
        if(self.results.has_key(candidateId)):            
            timeStart = self.results[candidateId].getTimeStart()
            return timeStart
        else:
            errorText = 'MultipleChoiceTest.getCandidateTimeStart()\nError: '\
                'no results for candidate with ID ' + candidateId
            log(errorText)
            return None
            
            
    def setCandidateTimeFinish(self, candidateId, dateTimeObject):
        """ Save the time when the candidate submitted his/her test. """
        #~ log('MultipleChoiceTest.setCandidateTimeFinish()\n')
        if(self.results.has_key(candidateId)):
            self.results[candidateId].setTimeFinish(dateTimeObject)
        else:
            errorText = 'MultipleChoiceTest.setCandidateTimeFinish()\nError: '\
                'no results for candidate with ID ' + candidateId
            log(errorText)
        # Make this un-undoable by the candidate
        makeTransactionUnundoable()
            
            
    def getCandidateTimeFinish(self, candidateId):
        """ Returns the time when the candidate submitted his/her test. """
        #~ log('MultipleChoiceTest.getCandidateTimeFinish()\n')
        if(self.results.has_key(candidateId)):
            timeFinish = self.results[candidateId].getTimeFinish()
            return timeFinish
        else:
            errorText = 'MultipleChoiceTest.getCandidateTimeFinish()\nError: '\
                'no results for candidate with ID ' + candidateId
            log(errorText)
            return None
            
    
    #~ security.declareProtected(PERMISSION_INTERROGATOR, 'getSubmitterIds')
    def getSubmitterIds(self):
        """ Return the IDs of the candidates who have actually submitted/taken this test. """
        return filter((lambda candidateId : self.hasSubmitted(candidateId)), self.results.keys())
            
            
    def getCandidateName(self, candidateId):
        """ Returns the name of the user with ID 'candidateId'. """
        # The name to be returned if an error occurred and the name could not be retrieved.
        errorVal = candidateId
        errorPrefix = "MultipleChoiceTest.getCandidateName("+str(candidateId) \
            + ") failed.\n\tError: "
        try:
            membershipTool = getToolByName(self, 'portal_membership')
            if (not membershipTool):
                log(errorPrefix + "getToolByName(self, 'portal_membership')" \
                    " returned None\n")
                return errorVal
            candidate = membershipTool.getMemberById(candidateId)
            if (not candidate):
                log(errorPrefix + "getMemberById() returned None\n")
                return errorVal
                
            # Try LDAP User Folder first
            sn = candidate.getProperty('sn', None)        # This should be configurable
            fn = candidate.getProperty('givenName', None) # This should be configurable
            if(sn and fn):
                candidateName = sn + ', ' + fn
            else:
                # Assume Plone User Folder
                candidateName = candidate.getProperty('fullname', None)
            if(not candidateName):
                log(errorPrefix + "getProperty('fullname', None) returned nothing\n")
                return errorVal
            return candidateName
        except:
            return errorVal
            
            
    #~ security.declareProtected(PERMISSION_INTERROGATOR, 'getResults')
    security.declarePrivate('setResults')
    # This function has to be implemented by classes derived from 
    # AbstractQuestionContainer
    def setResults(self, results = {}):
        """ Save the dictionary of CandidateTestResult objects in this test 
            instance. 
        """
        #~ log('MultipleChoiceTest.setResults()\n')
        if( type(results) != dict ):
            results = {}
        self.results = results
        #~ log('\tself.results.keys() = %s\n' % str(self.results.keys()))
        
    
    security.declareProtected(PERMISSION_INTERROGATOR, 'setEvaluationScripts')
    def setEvaluationScripts(self, evaluationScripts = {}):
        """ Save the dictionary of custom evaluation scripts in this 
            instance. 
        """
        if( type(evaluationScripts) == dict ):
            self.evaluationScripts = evaluationScripts
        else:
            evaluationScripts = {}
            
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'processEvaluationScriptUpload')
    def processEvaluationScriptUpload(self, archetype_name, funString):
        """ Routine for uploading custom evaluation scripts.
              The scripts are stored in the ObjectField 'evaluationScripts' of 
            the test instance. (That implies every test instance can have its 
            own evaluation script.)
            
            @param archetype_name The  archetype name of  the class you want  to customize,  e.g.
                    'Question' or 'MC Test'.
            @param funString A  string  containing a  python  script  that can  be  compiled with
                    Python's 'compile' function. The script has to define an evaluation function.
                    The name of that function has to be what is specified in the global constant 
                    CUSTOM_EVALUATION_FUNCTION_NAME  (see config.py).  See the last  line of the 
                    'getCandidatePoints()' methods of  'MultipleChoiceTest', 'QuestionGroup' and 
                    'Question' for the parameters these functions must accept.
        """
        #~ log("MultipleChoiceTest.processEvaluationScriptUpload(\n%s, \n%s)\n" \
        #~     %(str(archetype_name), funString))
        try: # Try to compile the script
            codeObj = compile(funString, '<string>', 'exec')
        except Exception, e:
            # Can't be compiled --> not useable
            errorText = str(e)
            return errorText
        noFunctionDefinition = True
        # Check if the script defines a function whose name is equal to the 
        # value of the constant 'CUSTOM_EVALUATION_FUNCTION_NAME'
        if(CUSTOM_EVALUATION_FUNCTION_NAME in codeObj.co_names):
            exec(codeObj)
            # Check if it is really a function and not a class or var
            if( getattr(eval(CUSTOM_EVALUATION_FUNCTION_NAME), 'func_name', None) \
                == CUSTOM_EVALUATION_FUNCTION_NAME ):
                noFunctionDefinition = False
        if(noFunctionDefinition):
            # Either nothing called 'CUSTOM_EVALUATION_FUNCTION_NAME' 
            # is defined or it's not a function
            return self.translate(\
                msgid   = 'uploadEvaluationScript_no_function_definition',\
                domain  = I18N_DOMAIN,\
                default = 'The script does not define a function called "%s"') \
                % CUSTOM_EVALUATION_FUNCTION_NAME
        # Save the script in the test's 'evaluationScripts' dictionary
        evaluationScripts = self.getEvaluationScripts()
        evaluationScripts[archetype_name] = funString
        self.setEvaluationScripts(evaluationScripts)
        #~ log("self.evaluationScripts = %s\n" %(str(self.getEvaluationScripts())))
        return None # return no error message (Success)
        
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'processQTIImport')
    def processQTIImport(self, file):
        """ Routine for uploading tests. 
        
            @param file A file containing a QTI assessmentItem or
                    a QTI package in a zip file.
        """
        addedObjects = []
        errorString = ''
        errors = MyStringIO()
        # try zipfile first
        zipFileInstance = None
        try:
            zipFileInstance = ZipFile(file.filename)
        except:
            try:
                zipFileInstance = ZipFile(file)
            except:
                pass
        if(zipFileInstance):
            # is a zip file
            addedObjects = importPackage(self, zipFileInstance, errors)
        else:
            # no zipfile --> maybe XML?
            # reset the file read ptr first
            try:
                file.seek(0)
            except:
                try:
                    file.filename.seek(0)
                except:
                    pass
            string = None
            try:
                string = file.filename.read()
            except:
                try:
                    string = file.read()
                except:
                    pass
            if(not (type(string) in [str, unicode])):
                # give up, no XML either
                # The file could not be read.
                errors.write( '\n' + self.translate(\
                    msgid   = 'file_read_error',\
                    domain  = I18N_DOMAIN,\
                    default = 'The file could not be read.') )
            else:
                # is an XML file
                addedObjects = importAssessmentItem(self, string, errors)
        # Reset the read ptr of the StringIO obj
        errors.seek(0)
        errorString = errors.read().strip()
        #~ log(errorString + '\n')
        return addedObjects, ([errorString, None][not errorString])
        
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'processQTIImport')
    def processQTIExport(self):
        """ Routine for downloading tests.
        """
        errors = MyStringIO()
        package = exportPackage(self, errors)
        # Reset the read ptr of the StringIO obj
        errors.seek(0)
        errorString = errors.read().strip()
        #~ log(errorString + '\n')
        return package, ([errorString, None][not errorString])
        
        
    security.declareProtected(PERMISSION_INTERROGATOR, 'getResultsAsList')
    def getResultsAsList(self, candidateIdList = None):
        """ Routine for exporting candidate results.
            
            @param candidateIdList A list of IDs of the candidates whose
                    results you want to know. 
                      If 'candidateIdList' is 'None', all results will be 
                    returned.
            
            @return A list of containing the results as a list of values. The values
                appear in the following order:
                
                [CandidateID, Candidate_Name, Score, Max. Score, 
                    Time Started, Time Finished]
                
                The first "result" ('getResultsAsList()[0]') is a header row 
                containing strings such as 'Candidate', 'Name', 'Score' etc.. 
                The actual results start at 'getResultsAsList()[1]'.
        """
        #~ log("MultipleChoiceTest.getResultsAsList()\n")
        """
            <th i18n:translate="candidate" 
                tal:content="python:headerRow[0]">Candidate</th>
            <th i18n:translate="name" 
                tal:content="python:headerRow[1]">Name</th>
            <th i18n:translate="test_results_score" 
                tal:content="python:headerRow[2]">Score</th>
            <th i18n:translate="test_results_max_score" 
                tal:content="python:headerRow[3]">Max. Score</th>
            <th i18n:translate="started" 
                tal:content="python:headerRow[4]">Started</th>
            <th i18n:translate="finished" 
                tal:content="python:headerRow[5]">Finished</th>
        """
        # The header row
        retVal = [[\
            self.translate(msgid = 'candidate', \
                      domain  = I18N_DOMAIN, \
                      default = 'Candidate'), \
            self.translate(msgid = 'name', \
                      domain  = I18N_DOMAIN, \
                      default = 'Name'), \
            self.translate(msgid = 'test_results_score', \
                      domain  = I18N_DOMAIN, \
                      default = 'Score'), \
            self.translate(msgid = 'test_results_max_score', \
                      domain  = I18N_DOMAIN, \
                      default = 'Max. Score'), \
            self.translate(msgid = 'started', \
                      domain  = I18N_DOMAIN, \
                      default = 'Started'), \
            self.translate(msgid = 'finished', \
                      domain  = I18N_DOMAIN, \
                      default = 'Finished')]]
        
        participants  = self.getParticipantIds()
        if(not (candidateIdList is None)):
            # If not all results are requested, filter 'candidateIdList' so 
            # that every 'candidateId' in 'candidateIdList' is actually a 
            # participant, i.e. is in 'self.getParticipantIds()'
            participants = filter((lambda candidateId : \
                (candidateId in participants)), candidateIdList)
        # We only want submitters who are also in the list of participants 
        # (i.e. in 'participants')
        submitters    = filter((lambda candidateId : \
            (candidateId in participants)), self.getSubmitterIds())
        # Participants who have not submitted their tests
        nonSubmitters = filter((lambda candidateId : \
            not (candidateId in submitters)), participants)
        # Sort 'participants' so that 'submitters' come first
        submitters.sort()
        nonSubmitters.sort()
        participants  = submitters + nonSubmitters
        results       = self.getResults()
        NaN           = '-' # Not a number/not available
        
        def formatTime(dateTimeObj):
            return localized_time( dateTimeObj, long_format = True, \
                context = self ) # 'localized_time' is a Plone function
        
        # The content rows
        for participantId in participants:
            points = None
            if(participantId in submitters):
                points       = self.getCandidatePoints(participantId)
                timeFinished = formatTime( self.getCandidateTimeFinish(\
                    participantId) )
            else:
                timeFinished = self.translate( msgid   = 'not_submitted', \
                                               domain  = I18N_DOMAIN, \
                                               default = 'Not submitted' )
            possiblePoints = self.getPossiblePoints(participantId)
            if(points is None):
                points = NaN
            if(possiblePoints is None):
                possiblePoints = NaN
            
            retVal.append( [\
                participantId, \
                self.getCandidateName(participantId), \
                points, \
                possiblePoints, \
                formatTime( self.getCandidateTimeStart(participantId) ), \
                timeFinished] )

        return retVal
            
            
    def getQuestionGroups(self):
        """ Returns all the QuestionGroup elements (no single questions). """
        return filter(lambda item : (item.Type() == QuestionGroup.archetype_name),\
            self.listFolderContents())
        
        
    def getPossiblePoints(self, candidateId):
        """ Return how many points the candidate could have got.
            
            @param candidateId the user ID of the candidate whose points you 
                    want to know. 
        """
        points = AbstractQuestionContainer.getPossiblePoints(self, candidateId)
        if(not isNumeric(points)):
            return None
        questionGroups = self.getQuestionGroups()
        for questionGroup in questionGroups:
            questionGroupPoints = questionGroup.getPossiblePoints(candidateId)
            if(not isNumeric(questionGroupPoints)):
                return None
            points = points + questionGroupPoints
        return points
            
            
    def getCandidatePoints(self, candidateId):
        """ Return how many points the candidate got for this test.
            
            @param candidateId the user ID of the candidate whose points you 
                    want to know.
            
            If a custom evaluation script has been uploaded it will be
            invoked. Otherwise a default method will be used. 
        """
        customScript = self.getEvaluationScript(self.archetype_name)
        questions = self.getQuestions(candidateId)
        questionGroups = self.getQuestionGroups()
        if(not customScript): # default
            points = 0
            for questionGroup in questionGroups:
                questionGroupCandidatePoints = \
                    questionGroup.getCandidatePoints(candidateId)
                if(not isNumeric(questionGroupCandidatePoints)):
                    return None
                points = points + questionGroupCandidatePoints
            for question in questions:
                if(not ('getCandidatePoints' in dir(question))):
                    return None
                questionCandidatePoints = \
                    question.getCandidatePoints(candidateId)
                if(not isNumeric(questionCandidatePoints)):
                    return None
                points = points + questionCandidatePoints
            return points
        else: # custom
            return evalFunString(customScript, CUSTOM_EVALUATION_FUNCTION_NAME, \
                [self, candidateId, questionGroups, questions])
            
                
    def getWorkflowState(self):
        """ Determine the Plone workflow state. """
        workflow_tool = getToolByName(self, 'portal_workflow')
        return workflow_tool.getInfoFor(self, 'review_state', None)
        
        
    # This function has to be implemented by classes derived from 
    # AbstractQuestionContainer
    security.declarePublic('isPublic')
    def isPublic(self):
        """ Determine whether this test has been published. """
        #~ log('MultipleChoiceTest.isPublic():\n')
        try:
            user = getSecurityManager().getUser()
            showUnpublishedContent = user.has_role('Manager') \
                and user.has_role('Authenticated')
            #~ log('\tshowUnpublishedContent = %s\n' %(str(showUnpublishedContent)))
            if showUnpublishedContent:
                return True
            if self.getWorkflowState() != 'published':
                #~ log("\t!'published'\n")
                return False
            now            = DateTime()
            startPublished = getattr(self, 'effective_date', None)
            endPublished   = getattr(self, 'expiration_date', None)
            #~ log('\tnow = ' + str(now) + '\n')
            #~ log('\tstartPublished = ' + str(startPublished) + '\n')
            #~ log('\tendPublished = ' + str(endPublished) + '\n')
            if((startPublished != None) and (startPublished > now)):
                #~ log("\tstartPublished > now\n")
                return False
            if((endPublished != None) and (now > endPublished)):
                #~ log("\tnow > endPublished\n")
                return False
            #~ log("\treturn True\n")
            return True
        except:
            #~ log("MultipleChoiceTest.isPublic() failed. An unknown exception occurred.\n")
            #~ log("\treturn False\n")
            return False        
        
        
    def deleteResultOf(self, candidateId):
        """ Deletes any saved information for candidate "candidateId", i.e.
            the test generated for him/her and his/her test results.
        """
        #~ log('MultipleChoiceTest.deleteResultOf(%s)\n' %(str(candidateId)))
        self.setResult(candidateId, None)
            
        
    def pre_validate(self, REQUEST, errors):
        """ This function is called when submitting the edit form (base_edit).
            
            @param REQUEST Contains the input for the form.
            @param errors If any validation error occurs, put it here.
                
            Requests from the 'base_edit' form to delete custom 
            evaluation scripts (from ObjectField 'evaluationScripts') 
            are processed here. 
        """
        if(REQUEST.get('deleteCustomEvaluationScripts', []) != []):
            scriptsToBeDeleted = REQUEST.get('deleteCustomEvaluationScripts', [])
            if( type(scriptsToBeDeleted) != list ):
                scriptsToBeDeleted = [scriptsToBeDeleted]
            evaluationScripts = self.getEvaluationScripts()
            for archetype_name in scriptsToBeDeleted:
                if(evaluationScripts.has_key(archetype_name)):
                    evaluationScripts.pop(archetype_name)
            self.setEvaluationScripts(evaluationScripts)
            
            
    #~ def post_validate(self, REQUEST, errors):
        #~ if(REQUEST.get('sourceFile_file', False)):
            #~ sourceFile = REQUEST.get('sourceFile_file')
            #~ try:
                #~ string = sourceFile.filename.read()
            #~ except:
                #~ string = sourceFile.read()
            #~ errorMsg = self.processQTIImport(string)
            #~ if(errorMsg):
                #~ errors['sourceFile'] = errorMsg + ' File=' + string
            #~ self.setSourceFile(None)
            
            
# Register this type in Zope
try:
    registerType(MultipleChoiceTest)
    log('Worked: registerType(MultipleChoiceTest)\n')
except Exception, e:
    log('Failed: registerType(MultipleChoiceTest): ' + unicode(e) + '\n')
    raise e
